use crate::config::BotConfig;
use crate::context::Context;
use chrono::Utc;
use log::{debug, info};
use poise::serenity_prelude::{CreateAttachment, MessageBuilder};
use poise::CreateReply;
use raas_types::raas::bot::roll::{roll_response, Roll, RollCmd};
use raas_types::raas::resp::response::Resp;
use raas_types::raas::service::raas_client::RaasClient;
use roll_rs::roll_inline;

type Error = Box<dyn std::error::Error + Send + Sync>;
type BotContext<'a> = poise::Context<'a, Context, Error>;

#[poise::command(slash_command, help_text_fn = "roll_help")]
async fn roll(
    ctx: BotContext<'_>,
    #[description = "Dice String eg 1d20"] roll: String,
) -> Result<(), Error> {
    let result = roll_inline(roll.as_str(), true)?;

    let mut msg = MessageBuilder::new();

    msg.push("Result: ");
    msg.push_bold_line_safe(result.dice_total.to_string());
    msg.push_mono_line_safe(result.string_result);

    ctx.reply(msg.build()).await?;

    Ok(())
}

fn roll_help() -> String {
    "Example Usage:\n`/roll d20+2`".to_string()
}

#[poise::command(slash_command, help_text_fn = "real_roll_help")]
async fn real_roll(ctx: BotContext<'_>) -> Result<(), Error> {
    let addr = ctx.data().config.raas_url.clone();

    let mut client = RaasClient::connect(addr).await?;

    let rolls = 3;

    let roll_request = raas_types::raas::cmd::Request {
        timestamp: Utc::now().timestamp() as u64,
        cmd: Some(raas_types::raas::cmd::request::Cmd::RollCmd(RollCmd {
            cmd: Some(raas_types::raas::bot::roll::roll_cmd::Cmd::Roll(Roll {
                rotations: rolls,
            })),
        })),
    };

    debug!("Request: {:?}", roll_request);

    let grpc_request = tonic::Request::new(roll_request);

    ctx.reply("Sending roll request, please hang on!").await?;

    let response = client.send_request(grpc_request).await?;

    let raas_response = response.into_inner();

    debug!(
        "Got RaaS: {} @ {}",
        raas_response.id, raas_response.timestamp
    );

    match raas_response.resp.unwrap() {
        Resp::RollResp(roll_resp) => {
            if let roll_response::Response::RollImage(img) = roll_resp.response.unwrap() {
                ctx.send(
                    CreateReply::default()
                        .content("Your roll:")
                        .attachment(CreateAttachment::bytes(img.img, "roll.jpg"))
                        .reply(true),
                )
                .await?;
            }
        }
        Resp::Error(err) => {
            ctx.reply(format!("Roll Bot encountered an error: `{}`", err.msg))
                .await?;
        }
    }

    Ok(())
}

fn real_roll_help() -> String {
    "Roll a real D20!".to_string()
}

pub async fn start_bot(config: BotConfig) {
    info!("Starting bot!");

    let intents = poise::serenity_prelude::GatewayIntents::non_privileged();

    let token = config.bot_token.clone();

    let framework = poise::Framework::builder()
        .options(poise::FrameworkOptions {
            commands: vec![roll(), real_roll()],
            ..Default::default()
        })
        .setup(|ctx, _ready, framework| {
            Box::pin(async move {
                poise::builtins::register_globally(ctx, &framework.options().commands).await?;
                Ok(Context { config })
            })
        })
        .build();

    let client = poise::serenity_prelude::ClientBuilder::new(token, intents)
        .framework(framework)
        .await;
    client.unwrap().start().await.unwrap();
}
